(***********************************************************************
   Delphi Unit: DelphiCAN.PAS (siehe auch TP-Unit MCAN)
   Standard CAN-I/O - Routinen
   bentigt: TVicHW32-Treiber

    J. Rathlev, IEAP, Uni-Kiel, (rathlev(a)physik.uni-kiel.de)

   The contents of this file may be used under the terms of the
   GNU Lesser General Public License Version 2 or later (the "LGPL")

   Software distributed under this License is distributed on an "AS IS" basis,
   WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
   the specific language governing rights and limitations under the License.

   ---------------------------------------------------------------
   untersttzte Hardware:
                  ISA-Interfacekarte,
                  Peak-Dongle mit LPT-Mux,
                  Peak-Dongle mit LPT-EPP,
                  IEAP-Dongle mit LPT-EPP,
                  Elektor-Dongle mit LPT-EPP

   Die Auswahl erfolgt beim Initialisieren (CANInit).

   Interrupt-Untersttzung:
   ------------------------
   Eingabe: Interrupt-Pufferung fr mehrere Identifier
            Fr jeden festen Identifier wird ein eigener Interruptbuffer
            unter einer Kanalnummer (2..8) angelegt (Routine NewCanID).
            Kanal 0 enthlt die Fehler (Interrupt + Status Reg.),
            Kanal 1 ist fr nicht spezifizierte ID reserviert (nur, wenn
              bei CanInit in Var. "Mode"  rcvAll gesetzt ist).
   Ausgabe: Interrupt-Pufferung

   Zeilenweise Ein- und Ausgabe von Text:
   --------------------------------------
   Allgemein: Eine Textzeile wird in Gruppen von max. 8 Bytes zerlegt.
            Eine CAN-Botschaft mit einer Lnge < 8 zeigt das Zeilenende
            an. Ist die Zeilenlnge ein Vielfaches von 8, wird am Ende
            eine Botschaft mit der Lnge 0 gesendet.
   Eingabe: fr jeden ID-Kanal wird ein Zhler (LCNT) mitgefhrt, der
            angibt, wieviele Zeilen sich im Puffer befinden.
            Der Zhler wird ber die Routine (CANLineCount) abgefragt.
            Das Einlesen einer Zeile erfolgt durch CANReadLn.
   Ausgabe: Die Ausgabe einer Textzeile erfolgt durch CANWriteLn.

   Objekte:
   --------
   - TCAN     : Basisklasse
   - TStdCAN  :


   J. Rathlev,   Feb. 1997
      Rev. 1,    Sep. 1997
      Rev. 2,    Okt. 1997  - neue Fehlerbehandlung
      Rev. 2.1,  Feb. 1998  - Fehler in CanInit behoben
      Rev. 2.2,  Apr. 1998  - Fehlerbehandlung verbessert
      Rev. 3,    Apr. 1998  - Neuorganisation der Routinen (Ra)

   Delphi5-Anpassung (J. Rathlev):
     Rev. 4,  Dez. 2001

   Neuorganisaton in eine Unit mit TVicHW-Treibern (J. Rathlev):
     Rev. 5,     Mai 2001
     Rev. 5.1,   Jan. 2004  - Anpassung an elektor-Dongle
     Rev. 5.2,   Aug. 2004  - Erweiterungsmglichkeit auf CAN 2.0
                              (siehe Var. CanProt in WriteCANBuffer und
                               ReadCANBuffer)
  *)

unit DelphiCan;

interface

uses WinTypes, WinProcs, SysUtils, Messages, TVicHW;

const
  (* Betriebsarten der Schnittstelle *)
  canISA = 1;     (* ISA-Karte *)
  canMux = 2;     (* Peak-Dongle mit LPT-Mux *)
  canEPP = 3;     (* Peak-Dongle mit LPT-EPP *)
  canIAP = 4;     (* IEAP-Dongle mit LPT-EPP *)
  canElk = 5;     (* Elektor-Dongle mit LPT-EPP *)
  canMax = canElk;

  (* Protokolle *)

  canVers1 = 1;   (* CAN 1.0 *)
  canVers2 = 2;   (* CAN 2.0 *)

  (* Kurzbezeichnungen der Betriebsarten *)
  CanModeShortList : array[1..canMax] of string[7] =
    ('ISA','normal','EPP','IEAP','Elektor');

  (* Bezeichnungen der Betriebsarten *)
  CanModeList : array[1..canMax] of string =
    ('CAN-Interface auf ISA-Karte',
     'Peak-Dongle an Standard-Parallelport',
     'Peak-Dongle an bidirektionalem Parallelport',
     'IEAP-Dongle an bidirektionalem Parallelport',
     'Elektor-Dongle an bidirektionalem Parallelport');

  (* Bezeichnungen der Protokolle *)
  CanProtList : array[1..2] of string =
     ('CAN 1.0','CAN 2.0');
         
  (* Empfangsmodus *)
  rcvAll = $80;  (* alle ID empfangen *)

  (* Standard-LPT *)
  LPTPortMax = 4;
  LPTPort: array [1..LPTPortMax] of word = ($378,$278,$3BC,$2BC);
  LPTIRQ : array [1..LPTPortMax] of byte = (7,5,3,9);

  {*** Baudratencodes = Registerwerte BTR0/BTR1 *}
  CANBaud1M   = $0014 ; {*   1 MBit / s *}
  CANBaud500K = $001C ; {* 500 kBit / s *}
  CANBaud250K = $011C ; {* 250 kBit / s *}
  CANBaud125K = $031C ; {* 125 kBit / s *}
  CANBaud100K = $432F ; {* 100 kBit / s *}
  CANBaud50K  = $472F ; {*  50 kBit / s *}
  CANBaud20K  = $532F ; {*  20 kBit / s *}
  CANBaud10K  = $672F ; {*  10 kBit / s *}
  CANBaud5K   = $7F7F ; {*   5 kBit / s *}

  CanBr5k = 1;
  CanBr10k = 2;
  CanBr20k = 3;
  CanBr50k = 4;
  CanBr100k = 5;
  CanBr125k = 6;
  CanBr250k = 7;
  CanBr500k = 8;
  CanBr1000k = 9;

  DefMaxrate = 9;
  BaudrateList : array[1..DefMaxrate] of string[11] =
      ('  5 kBit/s',' 10 kBit/s',' 20 kBit/s',' 50 kBit/s','100 kBit/s',
       '125 kBit/s','250 kBit/s','500 kBit/s','  1 MBit/s');

  BaudrateCodes : array[1..DefMaxrate] of word =
   (CANBaud5K,CANBaud10K,CANBaud20K,CANBaud50K,
    CANBaud100K,CANBaud125K,CANBaud250K,CANBaud500K,CANBaud1M);

  (* Control-Register: Enable Interrupts (error, transmit, receive) *)
  EnRxIRQ = 2;
  EnTxIRQ = 4;
  EnErIRQ = 8;
  EnOrIRQ = $10;
  EnableIRQ = EnRxIRQ+EnTxIRQ+EnErIRQ+EnOrIRQ;

  (* Interrupt Register (receive, transmit) *)
  RxInt = 1;
  TxInt = 2;

  {*** Maskenbits fuer Fehlerbedingungen ***}
  CANErrOK       = $00 ;
  CANErrXMTFULL  = $01 ; {* Sendepuffer im Controller ist voll *}
  CANErrRCVEMPTY = $02 ; {* keine neue Message empfangen *}
  CANErrOVERRUN  = $04 ; {* CAN-Controller wurde zu spaet gelesen *}
  CANErrBUSERROR = $08 ; {* Busfehler: ein Errorcounter erreichte Limit *}
  CANErrBUSOFF   = $10 ; {* Busfehler: CAN-Controller ging 'Bus-Off' *}
  CANErrChNr     = $20 ; {* ungltige Kanalnummer }
  CANErrBUFFULL  = $40 ; {* Schreibpuffer voll }

  (* ID-Kanal fr nicht spezifizierte ID *)
  DefCh = 1;

  (* Fehler bei CANInit *)
  errCANOK         = 0;    { kein Fehler beim Initialisieren }
  errCANRegTest    = 1 ;   { Fehler beim Registertest des CAN-Controllers }
  errCANIRQ        = 2 ;   { kein Interrupt spezifiziert }
  errCANChLimit    = 3 ;   { zu viele CAN-Kanle (max. IDMax) }
  errCANBaud       = 4 ;   { falsche Baud-Rate }
  errCANReset      = 5 ;   { Reset Error }
  errCANDevDrv     = 6 ;   { Fehler im TVicHW-Treiber }
  errCANLast       = errCANDevDrv;

// Init Fehler
  CanInitError : array[1..errCANLast] of string = (
    'Registertest failed','Illegal IRQ','Too much CAN channels',
    'Illegal baudrate','Reset error','TVicHW driver error');

// Betriebsfehler
  ErrCount = 7;
  CanErrors : array[1..ErrCount] of string =
    ('Transmit buffer overrun','Receive buffer empty',
     'Receiver overrun','Counter at limit',
     'Bus off state','Illegal channel','Tramsmit buffer full');

  (* ID-Kanal fr Fehler *)
  ErrCh = 0;

  { Windows-Botschaft bei CAN-Interrupt }
  wmCANReceived = wm_User+1;

  BufL  = 256;           (* Buffer Lnge *)
  IDMax = 8;             (* max. Anzahl v. Buffern *)

type
  (* Datenbuffer fr Can-Message *)
  TDataBuf = array[0..7] of byte;  (* CAN-Datenbuffer *)

  {* eine CAN-Message *}
  TCANMsg = record
    ID   : integer;   {* 11 Bit-Kennung                         *}
    RTR  : boolean ;  {* true, wenn remote request              *}
    LEN  : byte ;     {* Anzahl der gltigen Datenbytes (0..7)  *}
    IREG,             (* Interrupt Register 82C200 *)
    STAT : byte;      (* Status Register 82C200 *)
    Data : TDataBuf;  {* Datenbytes 0..7             *}
    end ;


  (* CAN-Parameter *)
  TCANPar = record
    Typ     : byte;      (* Interfacetyp (siehe CanIO) *)
    Port    : word;      (* Interfaceport *)
    IntNr   : byte;      (* Interruptnummer 3,4,5,7,10,11,.. *)
    Baud    : word;      (* 5,10,20,50,100,125,250,500,1000 in kBaud *)
    end;

  TBuffer = record
    ID,                     { Identifier }
    BCnt    : integer;      { buffer counter }
    UserCnt,                { user defined counter }
    BiPnt,                  { input buffer pointer }
    BoPnt   : word;         { output buffer pointer }
    Buffer  : array [0..BufL-1] of TCANMsg;  {input buffer }
    end;

  PTBuffer = ^TBuffer;

  ECanException = class (Exception)
  public
    ErrCode : word;
    constructor Create (CANErrNr : word);
    function ErrorMessage : string;
  end;

  TCanBuffer = record
    MainHandle  : HWnd;      { handle of main program }
    RcvAllCanID : boolean;   (* Standard: nur ausgewhlte ID empfangen *)
    FrstCh      : word;      { erster benutzbarer ID-Kanal }
    IDCnt       : word;      { hchste zulssige Kanalnummer }
    IBufPtr     : array [0..IDMax] of PTBuffer; { pointer to input data buffer }
    OBufPtr     : PTBuffer;      { pointer to output data buffer }
    end;
  PCanBuffer = ^TCanBuffer;

(* ----------------------------------------------------------------------- *)
(* CAN-Basisklasse:
   - Hardwareinitialisierung
   - Interruptbereitstellung (nachfolgende Klassen knnen eine Interruptroutine
     durch berschreiben der Methode "IntCallBack" bereitstellen
   - Bufferverwaltung
   - Lesen von Nachrichten aus dem Buffer
   - Senden von Nachrichten ber einen Buffer
   *)

  TCAN = class (TObject)
    constructor Create;
  protected
    FIrq           : word;
    FIntCallBack   : TUserCallBack;
    CanIOPtr       : PCanBuffer;
    NoMsgReg       : boolean ;  {* Flag fuer  'CANregisterMsg()' *}

    (* ----------------------------------------------------------------------- *)
    (* Hardware und Interrupt initialisieren *)
    procedure HwInit (Mode,PortAdr,IntNr  : word);

    (* Hardwareinterrupt zurcksetzen *)
    procedure HwRelease;

    (* Interrupt callback function *)
    procedure IntCallBack; virtual;

    (* Interrupt ein *)
    procedure EnableInt;

    (* Interrupt aus *)
    procedure DisableInt;

    (* ----------------------------------------------------------------------- *)
    (* Standardbuffer anlegen (bei Create-Methode aufrufen) *)
    procedure InitBaseBuffer;

    (* alle Standardbuffer entfernen (bei Destroy-Methode aufrufen)*)
    procedure ReleaseBaseBuffer;

    (* neuen CAN-Buffer anlegen *)
    procedure NewCANBuffer (Cnt,CID : word);

    (* CAN-Buffer entfernen *)
    procedure RemoveBuffer (Cnt : word);

    (* ----------------------------------------------------------------------- *)
    (* Empfangene Nachricht aus Buffer holen *)
    function  CANReadMsg (Nr        : word;
                          var CANMsg: TCANMsg) : word;

   (* Nachricht senden *)
    function  CANWriteMsg (const CANMsg : TCANMsg ) : word;

  public
    (* Neuen CAN-ID fr den Empfang eintragen *)
    function  NewCANID (var Nr : word;
                        CanID  : word) : word;

    (* CAN-Controller neu starten *)
    function CANRestart (btr0btr1 : word): word; virtual;

    (* RESET REQUEST lschen *)
    function CANReset : word; virtual;

    (* Status auslesen *)
    function ReadStatus : word;

    (* Warten bis CAN-Controller wieder aktiv ist *)
    function WaitCANReady : word;

    (* Warten bis CAN-Bus wieder aktiv ist *)
    function WaitBusReady : word;

    (* Nachrichtenfilter eintragen *)
    procedure CANRegisterMsg(ID: word);

    (*  Fehlernachricht generieren *)
    function CanErrorString (err : word; ErrDelim : char) : string;

    (* ----------------------------------------------------------------------- *)
    (* Routinen fr den Empfang von CAN-Nachrichten *)

    (* Anzahl der Botschaften im Empfangspuffer abfragen *)
    function  CANMsgCount (Nr : word) : word;

    (* ----------------------------------------------------------------------- *)
    (* Routinen fr das Senden von Nachrichten *)

    (* Anzahl der Botschaften im Sendepuffer abfragen *)
    function  CANXmtCount : word;

    (* Sende Buffer lschen *)
    function  CANClearSend : word;

  end;


(* -------------------------------------------------------------------------*)
(* Standard-CAN-Routinen *)
  TStdCAN = class (TCAN)
    constructor Create (Handle : HWnd);
    destructor Destroy; override;
  public
    (* CAN initialisieren *)
    procedure CANInit(Mode,PortAdr,IntNr,btr0btr1,Prot : word);

    (* CAN entlassen *)
    procedure CANClose ;

    (* Interrupt callback function *)
    procedure IntCallBack; override;

    (* -------------------------------------------------------------------------*)
    (* Eine empfangene Nachricht aus dem Buffer holen *)
    function  CANRead (Nr        : word;
                       var CANMsg: TCANMsg) : word;

    (* Eine Nachricht senden *)
    function  CANWrite (const CANMsg: TCANMsg) : word;

    (* -------------------------------------------------------------------------*)
    (* Zeilenweises Lesen und Schreiben ber CAN
       d.h. alle Nachrichten als Zeile (ASCII) interpretieren bis eine
       Nachricht weniger als 8 Datenbytes enthlt *)

    (* Anzahl der empfangenen Zeilen abfragen *)
    function  CANLineCount (Nr : word) : word;

    (* Empfangene Zeile aus Buffer holen *)
    function  CANReadLn (Nr        : word;
                        var Zeile : openstring) : word;
    (* Zeile senden *)
    function  CANWriteLn (CanID : word;
                          Zeile : string) : word;
  end;

(* -------------------------------------------------------------------------*)
(* -------------------------------------------------------------------------*)
implementation

(* ----------------------------------------------------------------------- *)
(* Fehlermeldungen generieren *)
(* CAN-Fehlermeldungen generieren *)
constructor ECanException.Create (CANErrNr : word);
begin
  ErrCode:=CanErrNr;
  inherited Create ('CAN Init: '+ErrorMessage);
  end;

function ECanException.ErrorMessage : string;
begin
  if ErrCode<=errCANLast then Result:=CANInitError[ErrCode]
  else Result:='unknown CAN error';
  end;

(* ----------------------------------------------------------------------- *)
const
  {**** Registeradressen des CAN_Controllers 82C200 ***}
  _CONTROL    =  0 ;
  _COMMAND    =  1 ;
  _STATUS     =  2 ;
  _INTERRUPT  =  3 ;
  _ACC_CODE   =  4 ;     {* Acceptance Code *}
  _ACC_MASK   =  5 ;     {* Acceptance Mask *}
  _TIMING0    =  6 ;     {* Bus Timing 0 *}
  _TIMING1    =  7 ;     {* Bus Timing 1 *}
  _OUTPUTCTRL =  8 ;     {* Output Control *}

  _XMTID      = 10 ;     {* Transmit Buffer: Identifier       *}
  _XMTLEN     = 11 ;     {*     RTR Bit und Data length Code  *}
  _XMTBUF     = 12 ;     {*     BYTE1 des data field's        *}

  _RCVID      = 20 ;     {* Receive Buffer: Identifier        *}
  _RCVLEN     = 21 ;     {*     RTR Bit und Data length Code  *}
  _RCVBUF     = 22 ;     {*     BYTE1 des data field's        *}

    (* Output Control: see data sheet for 82C200, Table 14, 15 *)
    OutPinPP0 = $18;  (* push/pull on TX0 *)
    OutPinPP1 = $C0;  (* push/pull on TX1 *)
    OutModeNorm = 2;  (* Normal Output, TX0, TX1: bit sequence *)

(* ----------------------------------------------------------------------- *)
(* ----------------------------------------------------------------------- *)
var
  OutPin,OutMode : byte;
  PortBase : word ;        (* Basisadresse deS Hardware-I/O-Ports *)
  CanMode  : word;         (* 1 = ISA, 2 = MUX, 3 = EPP, 4 = IEAP *)
  CanProt  : word;         (* 1 = CAN 1.0, 2 = CAN 2.0 *)

{************************************************************************}
{***    Dongle-Ansteuerung                                            ***}
{************************************************************************}
{    Allgemeines zur Ansteuerung bei Anschluss an LPT-Port:
     ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  Die Ansteuerung des PEAK-CAN-Dongles und des IEAP-CAN-Dongles
  erfolgt ber den LPT-Port.
  Die Signaladern des LPT-Portes sind ber 3 Register im IO-Bereich
  des PC's ansprechbar. Die Register liegen bei folgenden Adressen:
  "A" = Basisadresse      (Basisadr. z.B. $378 bei LPT1)
  "B" = Basisadresse+1
  "C" = Basisadresse+2

  Funktion der Registerbits:

  Signal          Richtung        Port               Stecker
  --------        ---------       ----------------   -------
  DATA0..DATA7    O (Basis-Mode)  "A"                2 .. 9
                  I/O (Bidi-Mode)

  STROBE*         O               "C", Bit 0         1
  ALF*            O               "C", Bit 1         14
  INIT*           O               "C", Bit 2         16
  DSL*            O               "C", Bit 3         17
  IRQ enable      (intern)        "C", Bit 4
  Read enable     (intern)        "C", Bit 5 (nur BIDI-Schnittstellen)

  ERR*            I               "B", Bit 3         15
  SEL             I               "B", Bit 4         13
  PEND            I               "B", Bit 5         12
  ACK*            I               "B", Bit 6 (IRQ)   10
  BUSY            I               "B", Bit 7         11

  Einige Port-C-Leitungen werden zustzlich invertiert,
  daher mssen Ausgaben nach Port C immer mit einem "XOR $0B"
  verrechnet werden.


  Konventionelle und bidirektionale LPT-Ports:
  ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  Im allgemeinen ist DATA0..7 des LPT-Port nicht bidirektional.
  Ausgaben an den CAN-Controllers gehen ber DATA0..7,
  Ergebnisse werden in 2 4-Bit-Nibbles ber Port "B" rckgelesen.
  (sogenannter "Multiplex-Modus").

  Wurde im PC-BIOS der LPT-Port auf "EPP" oder "EPC" gestellt,
  kann der LPT-Port ber das ECR-Register bei Basisadresse+$402
  in den PS/2-Modus geschaltet werden.
  Dann knnen Daten schneller ber DATA0..7 rckgelesen werden.


  Allgemeines zum CAN-Controller:
  ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  Der CAN-Controller 82C200 besitzt 32 interne Register.
  Diese sind 0..31 nummeriert.
  Ein Schreiben nach Register 31 lst einen Hardwarereset
  des 82C200 aus.


  Copyright (C) 1996 by PEAK Service GmbH, Darmstadt.

  Hinweise zum IEAP-Dongle:
  ~~~~~~~~~~~~~~~~~~~~~~~~
  Die Zuordnung der Steuersignale in Reg. PC ist im Schaltplan (siehe
  ELRAD 7/96 anders als im gekauften Peak-Dongle-Bausatz. Der IEAP-Dongle
  arbeitet ohne das PLD U3 und verwendet die Steuersignale wie im
  Schaltplan. Daher ist eine etwas andere Ansteuerung der Drucker-
  Schnittstelle erforderlich (Routinen: IEAPRegxxx)

  J. Rathlev, Aug. 1997
}
 { bergabeparameter fr alle nachfolgenden Routinen:
   'BasisAdr': Adresse des LPT-Ports (z.B. '$378' fr LPT1).
   'RegAdr':  anzusprechendes Register im 82C200
               (z.B. '2' fr das Statusregister).
   'RegVal':  in ein Register zu schreibender Wert.
 }

(* ----------------------------------------------------------------------- *)
(* MUX-Mode *)
CONST
  { -- Tabelle, die den an PB3..PB7 gelesenen Bits
       das originale Nibble am Ausgang des Multiplexers zuordnet.
       Besonderheit: PB6 wird ignoriert. }
  NibbleDecode : ARRAY[0..31] OF byte = (
        $08, $09, $0a, $0b, $0c, $0d, $0e, $0f,
        $08, $09, $0a, $0b, $0c, $0d, $0e, $0f,
        $00, $01, $02, $03, $04, $05, $06, $07,
        $00, $01, $02, $03, $04, $05, $06, $07 ) ;

procedure MuxRegInit (BasisAdr: word) ;
begin
  end;

function MuxRegRead (BasisAdr: word;
                     RegAdr  : byte) : byte;
var
  PA, PB, PC : word ;
  b0, b1 : byte ;
begin
  PA:=basisadr ; PB:=basisadr+1 ; PC:=basisadr+2 ;
  { -- Strobe (PC0) H->L: Adressuebernahme.
       Strobe L->H: Busfreigabe
       Mode (PC1) = L: MUX-Mode
       Nibbleumschaltung in der Lesephase mit PA6:
       PA6=0: PB3,PB4,PB5,PB7 = DATA4..7
       PA6=1:   "   "  "   "  = DATA0..3
  }
  { Bits in PC sind:  0 0 0 0 ispEN x Mode Strb Bemerkung}
  {                           ~~~~  ~ ~~~~~~~~~~~~~~~~~~~~~~~~}
  WritePort(PC,$0B xor $1D) ; { H     H L    H    Grundstellung }
  WritePort(PA,(RegAdr and $1F) or $80) ; { Bit 7 =H : Lesezyklus  }
  WritePort(PC,$0B xor $1C) ; { H     H L    L    Adr-Uebernahme }

  b1:=NibbleDecode[ReadPort(PB) shr 3] ; {     PA6=0: High Nibble lesen }
  WritePort(PA,$40) ;
  b0:=NibbleDecode[ReadPort(PB) shr 3] ; {     PA6=1: Low Nibble lesen }
  WritePort(PC,$0B xor $1D) ; { H     H L    H    Grundstellung }
  MuxRegRead:=(b1 shl 4) or b0 ;
  end;

procedure MuxRegWrite (BasisAdr: word;
                       RegAdr  : byte;
                       RegVal  : byte);
var
  PA, PC : word ;
begin
  PA:=basisadr; PC:=basisadr+2 ;
  { -- Strobe (PC0) H->L: Adressuebernahme.
       Strobe L->H: Busfreigabe
       Mode (PC1) = L: MUX-Mode }
  { Bits in PC sind:  0 0 0 0 ispEN x Mode Strb Bemerkung}
  {                           ~~~~  ~ ~~~~~~~~~~~~~~~~~~~~~~~~}
  WritePort(PC,$0B xor $1D) ; { H     H L    H    Grundstellung }
  WritePort(PA,RegAdr and $1F);{               Bit 7 =0 : Schreibzyklus  }
  WritePort(PC,$0B xor $1C) ; { H     H L    L    Adr-Uebernahme }
  WritePort(PA,RegVal) ;     { Schreibdaten ausgeben und ...   }
  WritePort(PC,$0B xor $1D) ; { H     H L    H    ... uebernehmen}
  {                                        Und wieder in Grundstellung }
  end;

(* ----------------------------------------------------------------------- *)
(* Standard-EPP-Mode *)
procedure EppRegInit (BasisAdr: word) ;
begin
 { 1284-kompatiblen Port in den "PS/2"-Mode setzen.
   Damit das ECR-Register bei Offset $402 freigeschaltet ist,
   mu im BIOS "EPP/EPC" eingestellt werden sein. }
  WritePort(basisadr+$402,(ReadPort(basisadr+$402) and $1F) or $20) ;
  end;

function EppRegRead (BasisAdr : word;
                     RegAdr   : byte) : byte;
var
  PA, PC : word ;
begin
  PA:=basisadr ; PC:=basisadr+2 ;
  { -- Strobe (PC0) H->L: Adressuebernahme.
       Strobe L->H: Busfreigabe
       Mode (PC1) = H EPP-Mode }
  { Bits in PC sind:  0 0 0 0 ispEN x Mode Strb Bemerkung}
  {                           ~~~~  ~ ~~~~~~~~~~~~~~~~~~~~~~~~}
  WritePort(PC,$0B xor $1F) ; { H     H H    H    Grundstellung }
  WritePort(PA,(RegAdr and $1F) or $80) ; { Bit 7 =H : Lesezyklus  }
  WritePort(PC,$0B xor $3E) ; { H     H H    L    Adr-Uebernahme, Tristate }
  EppRegRead:=ReadPort(PA) ;
  WritePort(PC,$0B xor $1F) ; { H     H H    H    Grundstellung }
  end;

procedure EppRegWrite (BasisAdr : word;
                       RegAdr   : byte ;
                       RegVal   : byte) ;
var
  PA, PC : word ;
begin
  PA:=basisadr ; PC:=basisadr+2 ;
  { -- Strobe (PC0) H->L: Adressuebernahme.
       Strobe L->H: Busfreigabe
       Mode (PC1) = H EPP-Mode }
  { Bits in PC sind:  0 0 0 0 ispEN x Mode Strb Bemerkung}
  {                           ~~~~  ~ ~~~~~~~~~~~~~~~~~~~~~~~~}
  WritePort(PC,$0B xor $1F) ; { H     H H    H    Grundstellung }
  WritePort(PA,RegAdr and $1F);{ Bit 7 =0 : Schreibzyklus  }
  WritePort(PC,$0B xor $1E) ; { H     H H    L    Adr-uebernahme }
  WritePort(PA,RegVal);      { Schreibdaten ausgeben und ...   }
  WritePort(PC,$0B xor $1F) ; { H     H H    H    ... uebernehmen}
  {                                             Und wieder in Grundstellung }
  end;

(* ----------------------------------------------------------------------- *)
(* IEAP-bidir-Mode - 82C200 Motorola-Mode *)
(* IEAP-Dongle zurcksetzen *)
procedure IapRegInit (BasisAdr : word);
var
  PC : word ;
begin
  (* set ECP-Control-Reg. to bidir. Mode *)
  WritePort(basisadr+$402,$20) ;
  PC:=basisadr+2;
  { Bits in PC sind:            4   3   2   1   0
                               IEN R/W  AS  E  RST}
  WritePort(PC,$0B xor $0D) ; { L   H   H   L    H    Grundstellung }
  WritePort(PC,$0B xor $0C) ; { L   H   H   L    L    Reset}
  WritePort(PC,$0B xor $0C) ; { L   H   H   L    L    Reset}
  WritePort(PC,$0B xor $11) ; { H   L   L   L    H    Grundstellung }
  end;

(* von IEAP-Dongle lesen *)
function IapRegRead (BasisAdr : word;
                     RegAdr   : byte) : byte;
var
  PA, PC : word ;
  b0     : byte;
begin
  PA:=basisadr; PC:=basisadr+2;
  { Bits in PC sind:            3   2   1   0
                               R/W  AS  E  RST}
  WritePort(PC,$0B xor $15) ; { L   H   L    H    Grundstellung }
  WritePort(PA,regadr and $1F);{ Bit 7 =0 : Schreibzyklus  }
  WritePort(PC,$0B xor $19) ; { H   L   L    H    Adr.uebernahme }
  WritePort(PC,$0B xor $3B) ; { H   L   H    H    Port auf Lesen }
  b0:=ReadPort(PA);                               { Daten lesen...   }
  WritePort(PC,$0B xor $19) ; { H   L   L    H    E aus }
  WritePort(PC,$0B xor $11) ; { L   L   L    H    Grundstellung }
  IapRegRead:=b0;
  end;

(* nach IEAP-Dongle schreiben *)
procedure IapRegWrite (BasisAdr      : word;
                       RegAdr,RegVal : byte) ;
var
  PA, PC : word ;
begin
  PA:=basisadr; PC:=basisadr+2;
  { Bits in PC sind:            3   2   1   0
                               R/W  AS  E  RST}
  WritePort(PC,$0B xor $15) ; { L   H   L    H    Grundstellung }
  WritePort(PA,regadr and $1F);{ Bit 7 =0 : Schreibzyklus  }
  WritePort(PC,$0B xor $11) ; { L   L   L    H    Adr.uebernahme }
  WritePort(PA,regval);      { Schreibdaten ausgeben und ...   }
  WritePort(PC,$0B xor $13) ; { L   L   H    H    ... uebernehmen}
  WritePort(PC,$0B xor $11) ; { L   L   L    H    Grundstellung }
  end;

(* ----------------------------------------------------------------------- *)
(* Elektor-bidir-Mode, 82C200: Intel-Mode *)
(* Elektor-Dongle initialisieren *)
procedure ElkRegInit (BasisAdr : word);
var
  PC : word ;
begin
  (* set ECP-Control-Reg. to bidir. Mode *)
  WritePort(basisadr+$402,$20) ;
  PC:=basisadr+2;
  { Bits in PC sind:            4   3   2   1   0
                               IEN ALE  RD  WR  - }
  WritePort(PC,$0B xor $17) ; { H   L   H   H   H    Grundstellung }
  end;

(* von Elektor-Dongle lesen *)
function ElkRegRead (BasisAdr : word;
                     RegAdr   : byte) : byte;
var
  PA, PC : word ;
  b0     : byte;
begin
  PA:=basisadr; PC:=basisadr+2;
  { Bits in PC sind:            3   2   1   0
                               ALE  RD  WR  - }
  WritePort(PC,$0B xor $1F) ; { H   H   H   H    Grundstellung }
  WritePort(PA,regadr and $1F);{ Bit 7 =0 : Schreibzyklus  }
  WritePort(PC,$0B xor $17) ; { L   H   H   H    Adr.uebernahme }
  WritePort(PC,$0B xor $33) ; { L   L   H   H    Port auf Lesen }
  b0:=ReadPort(PA);                              { Daten lesen...   }
  WritePort(PC,$0B xor $17) ; { L   H   H   H    E aus }
  ElkRegRead:=b0;
  end;

(* nach Elektor-Dongle schreiben *)
procedure ElkRegWrite (BasisAdr      : word;
                       RegAdr,RegVal : byte) ;
var
  PA, PC : word ;
begin
  PA:=basisadr; PC:=basisadr+2;
  { Bits in PC sind:            3   2   1   0
                               ALE  RD  WR  - }
  WritePort(PC,$0B xor $1F) ; { H   H   H   H    Grundstellung }
  WritePort(PA,regadr and $1F);{ Bit 7 =0 : Schreibzyklus  }
  WritePort(PC,$0B xor $17) ; { L   H   H   H    Adr.uebernahme }
  WritePort(PC,$0B xor $15) ; { L   H   L   H    Adr.uebernahme }
  WritePort(PA,regval);      { Schreibdaten ausgeben und ...   }
  WritePort(PC,$0B xor $17) ; { L   H   H   H    ... uebernehmen}
  end;

(* ----------------------------------------------------------------------- *)
(* Ansteuerung der ISA-Karte *)
function IsaRegRead (BasisAdr : word;
                     RegAdr   : byte) : byte;
begin
  IsaRegRead:=ReadPort(basisadr + RegAdr) ;
  end;

procedure IsaRegWrite (BasisAdr: word;
                       RegAdr  : byte;
                       RegVal  : byte) ;
begin
  WritePort(basisadr+RegAdr,RegVal);
  end;

(* ----------------------------------------------------------------------- *)
(* Can-I/O intialisieren *)
function InitCanIO (BasisAdr,Mode : word) : boolean;
begin
  PortBase:=BasisAdr;
  if Mode>canMax then Mode:=canIAP;
  CanMode:=Mode;
  OutPin:=OutPinPP0; OutMode:=OutModeNorm;
    case CanMode of
    canMux : MuxRegInit(PortBase);
    canEPP : EppRegInit(PortBase);
    canElk : ElkRegInit(PortBase);
    canIAP : IapRegInit(PortBase);
      end;
  Result:=true;   // Kompatibilitt zu Kithara
  end;

(* ----------------------------------------------------------------------- *)
(* Can-I/O reinitialisieren *)
procedure RestartCanIO;
begin
  if PortBase<>0 then begin
    case CanMode of
    canMux : MuxRegInit(PortBase);
    canEPP : EppRegInit(PortBase);
    canElk : ElkRegInit(PortBase);
    canIAP : IapRegInit(PortBase);
      end;
    end;
  end;

(* ----------------------------------------------------------------------- *)
(* Can-I/O Zustand *)
function CanIOActive : boolean;
begin
  CanIOActive:=PortBase<>0;
  end;

(* ----------------------------------------------------------------------- *)
(* Can-I/O abschalten *)
procedure ReleaseCanIO;
begin
  PortBase:=0;
  end;

(* ----------------------------------------------------------------------- *)
{*** aus einem Register des 82C200 lesen ***}
function ReadCanReg(RegAdr : byte) : byte ;
begin
  {* Auswahl der anzusteuernden CAN-Hardware: ISA-Einsteckkarte oder
   * eine der Donglevarianten *}
  case CanMode of
  canISA : ReadCanReg:=IsaRegRead(PortBase,regadr);
  canMux : ReadCanReg:=MuxRegRead(PortBase,regadr);
  canEPP : ReadCanReg:=EppRegRead(PortBase,regadr);
  canElk : ReadCanReg:=ElkRegRead(PortBase,regadr)
    else ReadCanReg:=IapRegRead(PortBase,regadr);
    end;
  end;

(* ----------------------------------------------------------------------- *)
{*** in ein Register des 82C200 schreiben ***}
procedure WriteCanReg (RegAdr : byte ;
                       RegVal : byte) ;
begin
  {* Auswahl der anzusteuernden CAN-Hardware: ISA-Einsteckkarte oder
   * eine der Donglevarianten *}
  case CanMode of
  canISA : IsaRegWrite(PortBase,regadr,regval);
  canMux : MuxRegWrite(PortBase,regadr,regval);
  canEPP : EppRegWrite(PortBase,regadr,regval);
  canElk : ElkRegWrite(PortBase,regadr,regval);
    else IapRegWrite(PortBase,regadr,regval);
    end;
  end;

(* ----------------------------------------------------------------------- *)
(* Registerwerte fr Baudrateneinstellung ermitteln *)
function CANBaud (Baud : word) : word;
begin
  case Baud of
  5    : CANBaud:=CANBaud5K;
  10   : CANBaud:=CANBaud10K;
  20   : CANBaud:=CANBaud20K;
  50   : CANBaud:=CANBaud50K;
  100  : CANBaud:=CANBaud100K;
  125  : CANBaud:=CANBaud125K;
  250  : CANBaud:=CANBaud250K;
  500  : CANBaud:=CANBaud500K;
  1000 : CANBaud:=CANBaud1M;
    else CANBaud:=0; (* Fehler *)
    end;
  end;

(* ----------------------------------------------------------------------- *)
(* Die Bits in Mask im Controllerregister RegAdr setzen *)
procedure CANBitSet (RegAdr : byte ;
                     Mask   : byte) ;
begin
  {* Besonderheit des COMMAND-Registers: Bits 4..0 sind write only *}
  if (RegAdr = _COMMAND )
    then WriteCanReg(_COMMAND, (ReadCanReg(_COMMAND) and $E0) or mask)
    else WriteCanReg(RegAdr, ReadCanReg(RegAdr) or mask) ;
  end;

(* ----------------------------------------------------------------------- *)
(* Die Bits in Mask im Controllerregister RegAdr loeschen *)
procedure CANBitClr (RegAdr : byte ;
                     Mask   : byte) ;
begin
  {* Besonderheit des COMMAND-Registers: Bits 4..0 sind write only *}
  if (RegAdr = _COMMAND )
    then WriteCanReg(_COMMAND, (ReadCanReg(_COMMAND) and $E0) and NOT mask)
    else WriteCanReg(RegAdr, ReadCanReg(RegAdr) and NOT mask) ;
  end;

(* ----------------------------------------------------------------------- *)
(* Sendebuffer des CAN-Controllers fllen *)
procedure WriteCANBuffer (const CANMsg : TCANMsg);
var
  i   : integer;
  tmp : byte;
begin
  {* Transmit Puffer fuellen *}
  with CANMsg do begin
    if CanProt=2 then begin  // CAN 2.0
//  hier Ansteuerung fr PeliCan-Mode (SJA1000) einfgen

      end
    else begin   // CAN 1.0
      WriteCanReg(_XMTID,(ID shr 3) and $ff) ;
      {* montiere das Laengenbyte der Message *}
      tmp:=((ID and $07) shl 5) or (LEN and $0f) ;
      {* remote-request-Message? *}
      if RTR then WriteCanReg( _XMTLEN, tmp or $10 )
      else begin
        {* sende Datenmessage *}
        WriteCanReg( _XMTLEN, tmp ) ;
        {* Bytes in den Sendepuffer des Controllers schreiben *}
        if LEN>0 then for i:=0 to pred(LEN) do
          WriteCanReg(_XMTBUF+i, DATA[i]) ;
        end;
      end;
    {* transmit data by setting of TRANSMISSION REQUEST *}
    end;
  CANBITSET(_COMMAND, $01) ;
  end;

(* ----------------------------------------------------------------------- *)
(* Empfangsbuffer des CAN-Controllers auslesen *)
procedure ReadCANBuffer (var CANMsg : TCANMsg);
var
  i   : integer;
  rcv : byte;
begin
  with CANMsg do begin
    if CanProt=2 then begin  // CAN 2.0
//  hier Ansteuerung fr PeliCan-Mode (SJA1000) einfgen

      end
    else begin // CAN 1.0
      {***  empfange jede Message, die dem Akzeptanzfilter genuegt *}
      rcv:=ReadCanReg(_RCVLEN);
      ID:=(word(ReadCanReg(_RCVID)) shl 3) or (rcv shr 5) ;
      LEN:=rcv and $0f ;
      STAT:=0; IREG:=0;
      if (rcv and $10 ) <> 0 then RTR:=true  {* RTR-Bit gesetzt! *}
      else begin
        {* Bytes aus dem Empfangspuffer holen *}
        RTR:=false ;
        if len>0 then for i:=0 to pred(LEN) do DATA[i]:=ReadCanReg(_RCVBUF+i) ;
        end { else } ;
        end;
    {* release receive buffer *}
    CANBITSET(_COMMAND, $04) ;
    end;
  end;

{*******************************************************************
 *      CANStatus
 *      Abfrage des Statusregisters des 82C200 und Rueckgabe
 *      einer oder mehrere CANErrxxx Bedingungen.
 *      *}
function  CANStatus : word;
var
  tmp: byte ;
  returnval: word;
begin
  returnval:=0 ;
  tmp:=ReadCanReg(_STATUS) ;
  if  (tmp and $01) = 0
    then returnval:=returnval or CANErrRCVEMPTY ;
  if ( tmp and $02 ) <> 0
    then returnval:=returnval or CANErrOVERRUN ;
  if ( tmp and $04 ) = 0
    then returnval:=returnval or CANErrXMTFULL ; {* transmit buffer locked *}
  if ( tmp and $20 ) <> 0
    then returnval:=returnval or CANErrXMTFULL ;
  if ( tmp and $40 ) <> 0
    then returnval:=returnval or CANErrBUSERROR;
  if ( tmp and $80 ) <> 0
    then returnval:=returnval or CANErrBUSOFF ;
  CANStatus:=returnval ;
  end; { "CANStatus" } 

(* ----------------------------------------------------------------------- *)
(* CAN-Object erzeugen und initalisieren *)
constructor TCAN.Create;
begin
  inherited Create;
  FIntCallBack:=nil; FIrq:=0;
  new(CanIOPtr);
  with CanIOPtr^do begin
    MainHandle:=0;
    end;
  end;

(* ----------------------------------------------------------------------- *)
(* Hardware und Interrupt initialisieren *)
procedure TCAN.HwInit (Mode,PortAdr,IntNr  : word);
begin
  if (not CanIOActive) then begin
    if InitCanIO (PortAdr,Mode) then begin
      FIrq:=IntNr; FintCallBack:=IntCallBack;
      EnableInterrupt (FIrq,FIntCallBack);
      end;
    end;
  end;

(* ----------------------------------------------------------------------- *)
(* Hardwareinterrupt zurcksetzen *)
procedure TCAN.HwRelease;
begin
  (* I/O, Interrupt und Callback zurcksetzen *)
  if CanIOActive then begin
    DisableInterrupt(FIrq);
    ReleaseCanIO;
    end;
  end;

(* -------------------------------------------------------------------------
   Interrupt callback routine*)
procedure TCAN.IntCallBack;
begin
  end;

(* ----------------------------------------------------------------------- *)
(* Interrupt ein *)
procedure TCAN.EnableInt;
begin
  EnableInterrupt (FIrq,FIntCallBack);
  end;

(* Interrupt aus *)
procedure TCAN.DisableInt;
begin
  DisableInterrupt (FIrq);
  end;


(* ----------------------------------------------------------------------- *)
(* Standardbuffer anlegen *)
procedure TCAN.InitBaseBuffer;
begin
  (* Standardbuffer fr Fehler ID anlegen *)
  with CanIOPtr^ do begin
    NewCANBuffer (ErrCh,0);
    IDCnt:=0;
    (* Sendebuffer anlegen *)
    New (OBufPtr);
    with OBufPtr^ do begin
      BCnt:=-1; UserCnt:=0; BoPnt:=0; BiPnt:=0; ID:=0;
      end;
    end;
  end;

(* ----------------------------------------------------------------------- *)
(* alle Standardbuffer entfernen *)
procedure TCAN.ReleaseBaseBuffer;
begin
  with CanIOPtr^ do begin
    dispose(OBufPtr);
    dispose(IBufPtr[ErrCh]);
    end;
  end;

(* ----------------------------------------------------------------------- *)
(* neuen CAN-Buffer anlegen *)
procedure TCAN.NewCANBuffer (Cnt,CID : word);
begin
  with CanIOPtr^ do begin
    New (IBufPtr[Cnt]);
    with IBufPtr[Cnt]^ do begin
      BCnt:=0; UserCnt:=0; BoPnt:=0; BiPnt:=0;
      ID:=CID;
      end;
    end;
  end;

(* ----------------------------------------------------------------------- *)
(* CAN-Buffer entfernen *)
procedure TCAN.RemoveBuffer (Cnt : word);
begin
  dispose(CanIOPtr^.IBufPtr[Cnt]);
  end;

(* ----------------------------------------------------------------------- *)
(* neuen Buffer fr einen ID anlegen
     Nr = Kanalnummer
     CanID = Identifier fr diesen Kanal *)
function TCAN.NewCANID (var Nr : word;
                        CanID  : word) : word;
begin
  with CanIOPtr^ do begin
    if IDCnt<IDMax then begin
      inc(IDCnt); Nr:=IDCnt;
      NewCanBuffer (IDCnt,CanID);
      NewCanID:=errCanOK;
      end
    else begin
      Nr:=0;
      NewCanID:=errCANChLimit;
      end;
    end;
  end;

(* ----------------------------------------------------------------------- *)
(* Die Routine programmiert das Empfangsfilter so um, da zustzlich zu
   allen anderen Nachrichts jetzt auch noch 'ID' empfangen wird.
   - Beim ersten Aufruf von 'CANRegisterMsg()' wird vom Empfang
     aller Nachrichts auf den Empfang eine einzigen umgeschaltet.
   - zurcksetzen durch spteres 'CANInit()' *)
procedure TCAN.CANRegisterMsg(ID: word ) ;
var
  n : byte ;
begin
  {* RESET, no Interrupts, Pin REF is Output,
   * Synch only by transitions from recessive to dominant  *}
  WriteCanReg(_CONTROL, $01) ;
  {* Algorithmus:
   * Acceptance-Code bleibt immer die beim 1. Aufruf angegebene ID.
   * Folgende Aufrufe loeschen nur noch Bits in der Acceptance-Mask.
   *}
  n := (ID shr 3) ;     {* nur die obersten 8 Bit der ID sind relevant *}
  if NoMsgReg then begin
        NoMsgReg := false ;
        WriteCanReg(_ACC_CODE, n) ;
        WriteCanReg(_ACC_MASK, 0) ; {* Alle Bits der ID sind relevant *}
      end else begin
        {* Check, in welchen Bits 'n' vom Acceptance-Code differiert.
         * Diese Bits zustzlich in der Acceptance-Mask setzen
         * (XOR ist 1 bei Differenz) *}
        n := n xor ReadCanReg(_ACC_CODE) ; {* n : high = differierende Bits *}
        n := n or ReadCanReg(_ACC_MASK) ; {* n : high = neue 'dont care' Bits *}
        WriteCanReg(_ACC_MASK, n) ;
      end { else } ;
  WriteCanReg(_CONTROL, $00) ;      {* Reset Ende *}
  end { "CANregisterMsg" } ;

(* ----------------------------------------------------------------------- *)
(* Warten bis CAN-Controller wieder aktiv ist *)
function TCAN.WaitCANReady : word;
var
  i : integer;
begin
  i:=0;
  if ReadCanReg(_CONTROL) and 1 <>0 then begin
    while (i<1000) and (ReadCanReg(_CONTROL) and 1 <>0) do begin
      inc(i); {delay(10);}
      end;
    end;
  if i>=10 then WaitCANReady:=errCANReset else WaitCANReady:=errCANOK;
  end;

(* ----------------------------------------------------------------------- *)
(* Warten bis CAN-Bus wieder aktiv ist *)
function TCAN.WaitBusReady : word;
var
  i : integer;
begin
  i:=0;
  if ReadCanReg(_STATUS) and $80 <>0 then begin
    while (i<10) and (ReadCanReg(_STATUS) and $80 <>0) do begin
      inc(i); {delay(10);}
      end;
    end;
  if i>=10 then WaitBusReady:=errCANReset else WaitBusReady:=errCANOK;
  end;

(* ----------------------------------------------------------------------- *)
(* CAN-Controller nach Fehler erneut starten *)
function TCAN.CANRestart (btr0btr1 : word): word;
begin
  RestartCanIO;
  WriteCanReg(_CONTROL,1) ;
  WriteCanReg(_COMMAND, $0E) ; {* CLr Overrun, Release RcvBuff, Abort Xmission *}
  WriteCanReg(_ACC_CODE, 0) ;
  WriteCanReg(_ACC_MASK, $ff) ;   {* Kein Bit der Kennung ist relevant *}
  WriteCanReg(_TIMING0, Hi(btr0btr1)) ;
  WriteCanReg(_TIMING1, btr0btr1 and $FF) ;
  NoMsgReg := true ;       {* Flag fuer  'CANRegisterMsg()' *}
  WriteCanReg(_OUTPUTCTRL, 0 or $18 or $02) ;
  WriteCanReg(_CONTROL,EnableIRQ) ;
  CanIOPtr^.OBufPtr^.BCnt:=-1;         (* reset send buffer *)
  CANRestart:=WaitCanReady;
  end;

(* ----------------------------------------------------------------------- *)
(* RESET REQUEST lschen (erforderlich nach BusOff-Fehler *)
function TCAN.CANReset : word;
begin
  WriteCanReg(_COMMAND,2) ;  (* abort transmission *)
  WriteCanReg(_CONTROL,EnableIRQ) ;
  CanIOPtr^.OBufPtr^.BCnt:=-1;         (* reset send buffer *)
  CANReset:=WaitBusReady;
  end;

(* ----------------------------------------------------------------------- *)
(* Status auslesen *)
function TCAN.ReadStatus : word;
begin
  Result:=CanStatus;
  end;
  
(* ----------------------------------------------------------------------- *)
(* Sende Buffer lschen *)
function TCAN.CANClearSend : word;
begin
  WriteCanReg(_COMMAND,2) ;  (* abort transmission *)
  CANClearSend:=CANStatus;
  end;

(* ----------------------------------------------------------------------- *)
(* Anzahl der Botschaften im Empfangspuffer des Kanals fr einen
   bestimmete ID abfragen *)
function TCAN.CANMsgCount (Nr : word) : word;
begin
  (* Buffer abfragen *)
  with CanIOPtr^do if Nr<=IDCnt then CanMsgCount:=IBufPtr[Nr]^.BCnt
  else CanMsgCount:=0;
  end;

(* -------------------------------------------------------------------------
  Fehlernachricht generieren *)
function TCan.CanErrorString (err : word; ErrDelim : char) : string;
var
  maske,i : word;
  s       : string;
begin
  maske:=1; i:=1; s:='';
  repeat
    if (err and maske)<>0 then begin
      if length(s)>0 then s:=s+ErrDelim+CanErrors[i]
      else s:=s+CanErrors[i];
      end;
    maske:=maske shl 1;
    inc(i);
    until i>ErrCount;
  if s='' then s:='Status = $'+IntToHex(err,4);
  Result:=s;
  end;

(* ----------------------------------------------------------------------- *)
(* Eine empfangene Nachricht aus dem Buffer holen
   Nr:     Empfangskanal (siehe NewCanID)
   CANMsg: Zeiger auf Nachrichtpuffer fuer die Rueckgabe der
           empfangenen Nachricht.
   Ergebnis: Fehlerstatus
     CANErrOK :      Kein Fehler
     CANErrRCVEMPTY: es wurde keine neue Nachricht empfangen.
     CANErrOVERRUN : eine empfangene Nachricht wurde nicht rechtzeitig
                     ausgelesen.
     CANErrBUSERROR: Controller meldet viele Busstoerungen.
     CANErrBUSOFF:   Controller ist im Bus-Off-Zustand und mu durch
                     CANReset wieder aktiviert werden.
     CANErrChNr :    falscher ID-Kanal *)
function TCAN.CANReadMsg (Nr        : word;
                          var CANMsg: TCANMsg) : word;
var
  returnval   : word;
begin
  {************************ Datenempfang **************************}
  returnval:=CANErrOK;
  (* aus Interrupt Buffer lesen *)
  with CanIOPtr^do if Nr<=IDCnt then with IBufPtr[Nr]^ do begin
    if BCnt>0 then begin
      CANMsg:=Buffer[BoPnt];
      dec(BCnt);
      inc(BoPnt); if BoPnt=BufL then BoPnt:=0;
      with CANMsg do begin
//        if ( Stat and $02 ) <> 0 then returnval := returnval or CANErrOVERRUN;
        if (Stat and $80)<>0 then returnval:=returnval or CANErrBUSOFF;
        if (Stat and $40)<>0 then returnval:=returnval or CANErrBUSERROR ;
        end;
      end
    else returnval:=CANErrRCVEMPTY ;
    end
  else returnval:=CANErrChNr;
  CANReadMsg:=returnval ;
  end;

(* ----------------------------------------------------------------------- *)
(* Anzahl der Botschaften im Sendepuffer abfragen *)
function TCAN.CANXmtCount : word;
begin
  (* Buffer abfragen *)
  CanXmtCount:=CanIOPtr^.OBufPtr^.BCnt
  end;

(* ----------------------------------------------------------------------- *)
(* Eine Nachricht senden.
   CANMsg: Zeiger auf Nachrichtpuffer fuer die Rueckgabe der
           empfangenen Nachricht.
   Ergebnis: Fehlerstatus
     CANErrOK     : Nachricht wurde erfolgreich abgesetzt.
     CANErrXMTFULL: Ausgabepuffer im CAN-Controller ist voll,
                    Nachricht wurde NICHT abgesetzt.
     CANErrBUSOFF:   Controller ist im Bus-Off-Zustand und mu durch
                     CANReset wieder aktiviert werden. *)
function  TCAN.CANWriteMsg (const CANMsg: TCANMsg) : word;
var
  returnval : word;
  Stat      : byte;
begin
  returnval := 0;
  with CanIOPtr^.OBufPtr^ do begin
    Stat:=ReadCanReg(_STATUS);
    if BCnt<0 then begin (* direkte Ausgabe *)
      {* Check 'Transmit Buffer Access'  *}
      {* transmit buffer locked ? *}
      if (Stat and $04) = 0 then returnval := CANErrXMTFULL
      else if (stat and $20 ) <> 0 then returnval := CANErrXMTFULL
        else begin
          WriteCANBuffer (CANMsg);
          BCnt:=0;
          end
      end
    else begin
      if (Stat and $80) <> 0 then returnval := CANErrBUSOFF
      else begin
        if BCnt<BufL then begin  (* in Buffer schreiben *)
          inc(BCnt);
          Buffer[BiPnt]:=CANMsg;
          inc(BiPnt); if BiPnt=BufL then BiPnt:=0;
          end
        else returnval := CANErrBUFFULL;
        end;
      end;
    end;
  CANWriteMsg := returnval ;
  end;

(* ----------------------------------------------------------------------- *)
(* Objekt anlegen, Buffer initialisieren *)
constructor TStdCAN.Create (Handle : HWnd);
begin
  inherited Create;
  with CanIOPtr^ do begin
    MainHandle:=Handle;
    (* Standard-Can-Buffer initialisieren *)
    InitBaseBuffer;
    FrstCh:=1; RcvAllCanID:=false;
    end;
  end;

(* ----------------------------------------------------------------------- *)
(* alle Can-Buffer entfernen *)
destructor TStdCAN.Destroy;
begin
  ReleaseBaseBuffer;
  inherited Destroy;
  end;

(* ----------------------------------------------------------------------- *)
(* Initialisieren der CAN-Schnittstelle und des Controllers
   Initialisiert den Interrupt,
   fuehrt einen Registertest auf dem 82C200 aus,
   programmiert dann Baudrate, Konfiguration der Sende/Empfangstreiber.
   Die Acceptance-Mask wird fuer den Empfang aller Messages gesetzt.
   Mode    : (siehe Unit CanIO: canXXX und rcvAll)
   PortAdr : Basisadresse der installierten Hadrware.
   IntNr   : Interruptnummer (z.B. 7 oder 5, 0 = polling)
   btr0btr1: Wert fuer die Timingregister des 82C200 (siehe Routine
             CANBaud in CanIO
   Prot    : Protokoll (CAN 1.0 oder 2.0)
   Ergebnis: Fehlercode (siehe DCanUtil) *)
procedure TStdCAN.CANInit(Mode,PortAdr,IntNr,btr0btr1,Prot : word);
var
  i    : integer;
  ok   : boolean;
begin
  if OpenDriver(0)then begin
    HwInit (Mode and $7F,PortAdr,IntNr);
    if btr0btr1=0 then raise ECANException.Create (errCANBaud)
    else with CanIOPtr^ do begin
      RcvAllCanID:=(Mode and rcvAll) <> 0;
      if RcvAllCanID then begin
        NewCANBuffer (DefCh,0);
        IDCnt:=DefCh; FrstCh:=succ(DefCh);
        end
      else begin
        IDCnt:=0; FrstCh:=1;
        end;
      {* RESET, Pin REF is Output,
       * Synch only by transitions from recessive to dominant  *}
      WriteCanReg(_CONTROL, $01) ;
      {* Register 4..8 sind 8 Bit read/write; 20..29 ist XMT-Buffer *}
      ok:=true;
      for i := 4 to 29 do if (i<9) or (i>19) then begin
        WriteCanReg(i,i+5) ; { jedes Register mit 'Adresse+5' testen }
        ok:=ok and (ReadCanReg(i)=i+5)
        end ;
      if ok then for i := 4 TO 29 do if (i<9) or (i>19) then ok:= ReadCanReg(i)=i+5;
      if ok then begin
        WriteCanReg(_COMMAND, $0E) ; {* CLr Overrun, Release RcvBuff, Abort Xmission *}
        {* Acceptance Code = Kennung fuer Empfang *}
        WriteCanReg(_ACC_CODE, 0) ;
        WriteCanReg(_ACC_MASK, $ff) ;   {* Kein Bit der Kennung ist relevant *}
        NoMsgReg:= true ;        {* Flag fuer  'CANRegisterMsg()' *}
        {* TIMING0/1 fuer versch. ISO-Baudraten bei 16Mhz Quarz
         * [PLC] S.4, ausfuehrlich [TDS] S. 46f   *}
        WriteCanReg(_TIMING0, Hi(btr0btr1)) ;
        WriteCanReg(_TIMING1, btr0btr1 and $FF) ;

        {* Output Control: Push/Pull on TX0, TX1 disabled
         Elektor-Dogle:
           * OCTP1 = OCTN1 = 1, OCPOL1 = 0 (keine Invertierung)
           * OCTP0 = OCTN0 = OCPOL0 = 0
         andere:
           * OCTP1 = OCTN1 = OCPOL1 = 0,
           * OCTP0 = OCTN0 = 1, OCPOL0 = 0 (keine Invertierung)
         * OCMODE1 = 1, OCMODE0 = 0  (normal)
         * [TDS] S.50 u. 119 *}
        if CanMode=canElk then WriteCanReg(_OUTPUTCTRL, $C0 or 0 or $02)
        else WriteCanReg(_OUTPUTCTRL, 0 or $18 or $02) ;
        (* Clear reset-request, enable interrupts *)
        WriteCanReg(_CONTROL,EnableIRQ) ;
        i:=WaitCanReady;
        if i<>errCanOK then raise ECANException.Create (i);
        end
      else raise ECANException.Create (errCANRegTest);
      end
    end
  else raise ECANException.Create (errCANDevDrv);
  end;

(* ----------------------------------------------------------------------- *)
(* Controller deaktivieren *)
procedure TStdCAN.CANClose ;
var
  i    : integer;
begin
  if CanIOActive then WriteCanReg(_CONTROL, $01) ; {* Controller in Reset-Mode *}
  HwRelease;
  with CanIOPtr^ do begin
    rcvAllCanID:=false;
    for i:=DefCh to IDCnt do RemoveBuffer (i);
    IDCnt:=0;
    end;
  CloseDriver;
  end;

(* -------------------------------------------------------------------------
   Interrupt callback routine *)
procedure TStdCAN.IntCallBack;
var
  IR  : byte;
  Msg : TCanMsg;
  i   : integer;

  (* CAN-Botschaft in Puffer schreiben *)
  procedure WriteInputBuffer (n       : integer;
                              var msg : TCanMsg);
  begin
    with CanIOPtr^ do if n<=IDCnt then with IBufPtr[n]^ do begin
      if BCnt<BufL then begin
        Buffer[BiPnt]:=msg;
        if msg.Len<8 then inc(UserCnt);
        inc(BCnt); inc(BiPnt);
        if BiPnt=BufL then BiPnt:=0;
        end;
      end;
    end;

begin
  with CanIOPtr^ do begin
    IR:=ReadCanReg(_INTERRUPT) and $1f;
    if (IR and TxInt) <> 0 then with OBufPtr^ do repeat (* Senden *)
      dec(BCnt);
      if BCnt>=0 then begin
        WriteCANBuffer (Buffer[BoPnt]);
        inc(BoPnt); if BoPnt=BufL then BoPnt:=0;
        end;
      until (BCnt<0) or (ReadCanReg(_STATUS) and $24<>4);
    if( IR and not TxInt ) <> 0 then begin (* Error oder Empfang *)
      (* Post message for main program *)
      PostMessage (MainHandle,wmCANReceived,IR,0);

      if ( IR and RxInt ) <> 0 then begin (* Empfangen *)
        ReadCANBuffer (msg);
        (* ID vergleichen *)
        i:=FrstCh;
        with msg do while (i<=IDCnt) and (ID<>IBufPtr[i]^.ID) do inc(i);
        if RcvAllCanID and (i>IDCnt) then i:=DefCh;
        with msg do begin
          IREG:=IR; STAT:=ReadCanReg(_STATUS);
          end;
        end
      else begin  (* Error interrupts *)
        with msg do begin
          ID:=$7FFF; RTR:=false;
          LEN:=0; i:=0;
          IREG:=IR; STAT:=ReadCanReg(_STATUS);
          end;
        end;
      WriteInputBuffer(i,msg);
      with msg do begin (* Error handling *)
        (* Overrunstatus lschen *)
        if ( Stat and $02 ) <> 0 then CANBITSET(_COMMAND, $08 ) ;
        (*** Bus-Off: Controller is NOT involved in bus activities.
          Der Controller geht wieder 'On-Bus' nach dem Empfang
          von 128 * 11 aufeinanderfolgenden rezessiven Bits. *)
        if ( Stat  and $80 ) <> 0 then CANBITCLR(_CONTROL, $01) ;
        end;
      end;
    end;
  end;


(* ----------------------------------------------------------------------- *)
(* Anzahl der Textzeilen im Empfangspuffer des Kanals fr einen
   bestimmte ID abfragen *)
function TStdCAN.CANLineCount (Nr : word) : word;
begin
  (* Buffer abfragen *)
    with CanIOPtr^ do if Nr<=IDCnt then CanLineCount:=IBufPtr[Nr]^.UserCnt
  else CanLineCount:=0;
  end;

(* ----------------------------------------------------------------------- *)
(* Eine empfangene Nachricht aus dem Buffer holen
   Zeilenzhler berwachen, sonst wie CAnReadMsg (Unit CanUtil)
   Ergebnis: Fehlerstatus (siehe CanIO) *)
function  TStdCAN.CANRead (Nr        : word;
                           var CANMsg: TCANMsg) : word;
begin
  Result := CANErrOK ;
  (* aus Interrupt Buffer lesen *)
  with CanIOPtr^ do if Nr<=IDCnt then with IBufPtr[Nr]^ do begin
    if BCnt>0 then begin
      Result:=CANReadMsg (Nr,CANMsg);
      if (CANMsg.Len<8) and (UserCnt>0) then dec(UserCnt);
      end
    else Result := CANErrRCVEMPTY ;
    end
  else Result:=CANErrChNr;
  end;

(* ----------------------------------------------------------------------- *)
(* Textzeile einlesen, ev. warten bis vollstndig
   Nr:     Empfangskanal (siehe NewCanID)
   Zeile:  Empfangene Textzeile (Lnge<=255)
   Ergebnis: Fehlerstatus (siehe CANRead) *)
function TStdCAN.CANReadLn (Nr        : word;
                          var Zeile : openstring) : word;
var
  msg    : TCanMsg;
  ErrStat,
  i,stat : word;
  ende   : boolean;
begin
  Zeile:=''; ende:=false; ErrStat:=0;
  repeat
    stat:=CANRead(Nr,msg);
    if (stat and CANErrRCVEMPTY)=0 then with msg do begin
      if (Len>1) or (Data[0]<>0) then  (* nur wenn kain Dummy *)
        for i:=0 to pred(Len) do Zeile:=Zeile+chr(Data[i]);
      ende:=msg.len<8;
      end;
    ErrStat:=ErrStat or (stat and not CANErrRCVEMPTY);
    until ende;
  CANReadLn:=ErrStat;
  end;

(* ----------------------------------------------------------------------- *)
(* Eine Nachricht senden
   identisch mit CANWriteMsg
   Ergebnis: Fehlerstatus *)
function TStdCAN.CANWrite (const CANMsg: TCANMsg) : word;
begin
  CANWrite:=CANWriteMsg (CANMsg);
  end;

(* ----------------------------------------------------------------------- *)
(* Textzeile senden
   Zeile:  Zu sendende Textzeile (Lnge<=255)
   Ergebnis: Fehlerstatus (siehe CANRead) *)
function TStdCAN.CANWriteLn (CanID : word;
                           Zeile : string) : word;
var
  msg      : TCanMsg;
  ErrStat,n,
  i,j      : integer;
begin
  ErrStat:=0; msg.ID:=CanID;
  n:=length(Zeile) div 8;
  (* prfen, ob genug Platz im Puffer ist *)
    with CanIOPtr^ do if OBufPtr^.BCnt+n<BufL then begin
    if n>0 then for i:=0 to pred(n) do begin
      with msg do begin
        Len:=8; RTR:=false;
        for j:=1 to 8 do Data[pred(j)]:=ord(Zeile[8*i+j]);
        end;
      ErrStat:=ErrStat or CANWrite (msg);
      end;
    with msg do begin
      Len:=length(Zeile) mod 8; RTR:=false;
      if Len=0 then begin
        inc(Len); Data[0]:=0; (* dummy, um Len=0 auszuschlieen *)
        end
      else for j:=1 to Len do Data[pred(j)]:=ord(Zeile[8*n+j]);
      end;
    CANWriteLn:=ErrStat or CANWrite (msg);
    end
  else begin
    CANWriteLn:=CANErrBUFFULL or CANStatus;
{    if CANStatus and CANErrBUSOFF <>0 then begin
      CANReset; OBufPtr^.BCnt:=-1; (* Buffer lschen *)
      end;}
    end;
  end;

begin
  PortBase:=0; CanMode:=canIAP; CanProt:=canVers1;
  end.


