(* Interface to TVicHW32 drivers

   © J. Rathlev, IEAP, Uni-Kiel, (rathlev(a)physik.uni-kiel.de)

   The contents of this file may be used under the terms of the
   GNU Lesser General Public License Version 2 or later (the "LGPL")

   Software distributed under this License is distributed on an "AS IS" basis,
   WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
   the specific language governing rights and limitations under the License.

   Requirements:
     Window 95/98/ME:
       - "tvichw32.vxd" must be in "<windows>\system"
       - "tvichw32.dll" can be in "<windows>\system" or in the folder from where
           you start the application using TVicHW32

     Windws NT/2000/XP:
       - "tvichw32.sys" must be in "<windows>\system32\drivers"
           (you need administrator rights to copy the file to this place)
       - "tvichw32.dll" can be in "<windows>\system32" or in the folder from where
           you start the application using TVicHW32
       - The following registry keys must be set
          [HKEY_LOCAL_MACHINE\SYSTEM\CurrentControlSet\Services\TVicHW32]
          "ErrorControl"=dword:00000001
          "Type"=dword:00000001
          "Start"=dword:00000002
          "Group"="Extended Base"
          "Parameters"=""
          "DisplayName"="TVicHW32"

   Enhancements:
     TvicLib has three compiler options (look there):
     1. Static load of TVicHW32.DLL
     2. Dynamic load automatically on startup
     3. Dynamic load by user
     Dynamic load by user gives then possibilty to check the return value
     of function "LoadDriver" for the availability of the drivers.

   J. Rathlev, April 2002;
   *)

unit TVicHW;

interface

uses Windows, SysUtils, TVicLibDyn, Hw_Types;

type
  TAccessMode = (amSoft,amHard);

  TUserCallBack = procedure of object;

(* -----------------------------------------------------------------------
  Interface to TVicHW32 drivers
  see: TVicHW32.chm (www.entechtaiwan.com/tools.htm)  *)
function LoadDriver : boolean;
function UnLoadDriver : boolean;

(* ----------------------------------------------------------------------- *)
(* Initialize TVicHW32 drivers *)
function OpenDriver (EntryPoint : word) : boolean;
procedure CloseDriver;
function DriverActive : boolean;

(* ----------------------------------------------------------------------- *)
(* Interrupts *)
procedure EnableInterrupt (Irq       : word;
                           hCallback : TUserCallBack);
procedure DisableInterrupt(Irq : word);

(* ----------------------------------------------------------------------- *)
(* I/O *)
procedure InitPortIO (AccessMode : TAccessMode);
function ReadPort (Port : word) : byte;
procedure WritePort (Port : word; val : byte);
function ReadPortW (Port : word) : word;
procedure WritePortW (Port : word; val : word);

(* ----------------------------------------------------------------------- *)
(* Get HDD info *)
function ReadHDDInfo (Drive,Master : integer) : THDDInfo;

(* ----------------------------------------------------------------------- *)
implementation

var
  UserCallBack : TUserCallBack;
  Hw32         : THANDLE;
  ActiveHW     : Boolean;

(* ----------------------------------------------------------------------- *)
(* Load TVicHW32 drivers *)
function LoadDriver : boolean;
begin
  Result:=LoadTVicHW32;
  end;

(* Unload TVicHW32 drivers *)
function UnLoadDriver : boolean;
begin
  Result:=UnLoadTVicHW32;
  end;

(* ----------------------------------------------------------------------- *)
(* Initialize TVicHW32 drivers *)
function OpenDriver (EntryPoint : word) : boolean;
begin
  UserCallBack:=nil;
  HW32:=0;
  if EntryPoint <2 then begin
    HW32:=OpenTVicHW32(HW32,'TVicHW32',pchar('TVicDevice'+IntToStr(EntryPoint)));
    ActiveHW:=GetActiveHW(HW32);
    Result:=ActiveHW;
    end
  else Result:=false;
  end;

(* Close TVicHW32 drivers *)
procedure CloseDriver;
begin
  HW32:=CloseTVicHW32(HW32);
  ActiveHW:=FALSE;
  end;

(* Check if active *)
function DriverActive : boolean;
begin
  Result:=GetActiveHW(HW32);
  end;

(* ----------------------------------------------------------------------- *)
(* Interface to Callback function *)
procedure CallBack (ActIrq : word); stdcall;
begin
  if assigned(UserCallBack ) then UserCallBack;
  end;

(* ----------------------------------------------------------------------- *)
(* Enable interrupt *)
procedure EnableInterrupt (Irq       : word;
                           hCallback : TUserCallBack);
begin
  UserCallBack :=hCallback;
  UnmaskIRQ(HW32,Irq,@CallBack);
  end;

(* Disable interrupt *)
procedure DisableInterrupt(Irq : word);
begin
  MaskIRQ(HW32,Irq);
  end;

(* ----------------------------------------------------------------------- *)
(* Initialize I/O *)
procedure InitPortIO (AccessMode : TAccessMode);
begin
  SetHardAccess(Hw32,AccessMode=amHard);
  end;

(* ----------------------------------------------------------------------- *)
(* Port I/O using TVicHW32 driver *)
(* replace b:=Port[adr] by b:=ReadPort(Adr) *)
function ReadPort (Port : word) : byte;
begin
  Result:=GetPortByte(Hw32,Port);
  end;

(* replace Port[adr]:=b by WritePort(Adr,b) *)
procedure WritePort (Port : word; val : byte);
begin
  SetPortByte(Hw32,Port,val);
  end;

(* replace w:=Portw[adr] by b:=ReadPortW(Adr) *)
function ReadPortW (Port : word) : word;
begin
  Result:=GetPortWord(Hw32,Port);
  end;

(* replace Portw[adr]:=b by WritePortW(Adr,b) *)
procedure WritePortW (Port : word; val : word);
begin
  SetPortWord(Hw32,Port,val);
  end;

(* ----------------------------------------------------------------------- *)
(* Get HDD info *)
function ReadHDDInfo (Drive,Master : integer) : THDDInfo;
begin
  GetHDDInfo (Hw32,Drive,Master,@Result);
  end;

end.
